from . import segment
from delune import _delune
import types

class SegmentWriter (segment.Segment):
	def __init__ (self, si):	
		self.si = si
		newseg = self.si.getNewSegment ()		
		segment.Segment.__init__ (self, None, -1, 'w', version = self.si.version)
		self.memUsage = 4096 #default document buffer
		self.termTable = _delune.TermTable ()
	
	def getMemUsage (self):	
		return self.memUsage + self.fd.usage ()
		
	def newdocid (self):
		if not self.opened:
			newseg = self.si.getNewSegment ()
			self.open (self.si.fs.new (newseg), newseg)
		
		docId = self.numDoc
		self.numDoc += 1
		return docId		
		
	def close (self):
		if self.termTable:
			self.termTable.close ()
			self.termTable = None
			
		segment.Segment.close (self)
			
	def writeField (self, fields, document = ""):
		self.fd.write (fields, document)
	
	def writeSortKey (self, fieldNum, fieldValue, size):
		if type (fieldValue) is list:
			self.memUsage += self.sm.addIntList (fieldNum, fieldValue, size) * size
		else:
			self.memUsage += self.sm.addInt (fieldNum, fieldValue, size) * size
	
	def writeStringSortKey (self, fieldNum, fieldValue, size):
		self.memUsage += self.sm.addShortString (fieldNum, fieldValue, size) * size
	
	def writeNorm (self, fieldNum, numTerm):
		self.memUsage += self.sm.addNorm (fieldNum, numTerm)
		
	def writePosting (self, docId, fieldNum, fieldValue, tf, pos = None):
		if len (fieldValue) > 100:
			self.si.log ("field `%d-%s` is over 100 chars, forcely truncated." % (fieldNum, fieldValue), "warn")
			fieldValue = fieldValue [:100]
		self.memUsage += self.termTable.add (fieldValue, fieldNum, docId, tf, pos)
	
	def flush (self):
		if not self.opened: return
		self.si.log ("writing %d documents to segment %s..." % (self.numDoc, self.seg))
		
		for term, fdno, posting in self.termTable:
			frqposition, prxposition = self.tf.tell ()			
			df, skip, pskip = self.tf.write (posting)
			self.ti.add (term, fdno, df, frqposition, prxposition, skip - frqposition, pskip - prxposition)
			self.tf.commit ()
		
		self.si.log ("flushed %d terms..." % self.termTable.numterm)	
		self.si.addSegment (self.seg, self.numDoc)
		self.si.log ("closing segment...")	
		self.close ()
		self.si.flush ()		
		self.si.log ("flushing done.")		
		
		


from . import naivebayes
import math

class Classifier (naivebayes.Classifier):
	
	def setopt (self, **opts):		
		self.fetchcount = int (opts.get ("topdoc", 100))
		naivebayes.Classifier.setopt (self, **opts)
	
	def getopt (self, name):
		if name == "fetch":
			return self.fetchcount
		
	def match (self, name):
		return name in ("tfidf",)
	
	def query (self, terms, surfix, fetch = 10, lang = "un"):
		qs = "%s (%s)" % (surfix and "(%s)" % surfix or "", " or ".join ([x [0] for x in terms]))
		res = self.reader.searcher.do_query (
			qs, 0, self.fetchcount == 0 and 100000000 or self.fetchcount, 
			sort = self.fetchcount > 0 and "tfidf" or "", # self.fetchcount == -1, then fetch all
			lang = lang,
			analyze = False, limit = False
		)
		return res
		
	def guess (self, mem, qs, lang = "un", cond = ""):
		terms = self.getFeatures (mem, qs, lang)
		if not terms: return []
					
		res = self.query (terms, cond, self.fetchcount, lang)
		if res ["total"] == 0:
			return []
				
		result = []
		for row  in res ["result"]:
			cla, score = row [0][0], row [-1]
			result.append ((cla, score))		
		return result



import os
import sys
import types
import delune
from delune import _delune
from delune.searcher import searcher
from delune.searcher.segment import segmentreader
from . import segment
import math
from delune import memory

class ComposedSegmentReader:
	def __init__ (self, si):
		self.si = si
		self.logger = self.si.logger
		
		self.reader = None
		self.searcher = None
		self.reader = None
		self.corpus = None
		
		self.IDFCache = {}
		self.DFCache = {}
		self.termScoreCache = {}		
		self.numpools = len (self.si.segments.pools)
	
	def getSegmentNum (self):
		return self.reader.seg
	
	def translate (self, res):
		res = [(self.si.segments.pools [x [0]], x [1]) for x in res]
		res.sort (key = lambda x: x [1], reverse = True)
		return res
	
	def get_memory (self):
		return memory.get (self.reader)
					
	def load (self):
		if self.reader:
			self.close ()
		
		self.si.read ()
		seg = self.si.getSegmentList ()[0]
		
		self.reader = segment.Segment (self.si.fs.get (seg), seg, "r", version = self.si.version)
		self.reader.open ()
		self.reader.mutex_id = memory.set (self.reader)
		
		df, doff, poff, skip, plen = self.reader.ti.get (memory.get (self.reader), "corpus", 0)
		self.corpus = [x [1:] for x in self.reader.tf.get (memory.get (self.reader), df, doff, 3)]
		col = delune.collection (os.path.join (self.si.fs.getmaster (), "%s.ts" % self.si.alias), delune.READ, self.si.analyzer, self.si.logger, version = self.si.version)		
		self.searcher = col.get_searcher ()
		self.fdnoterm = self.searcher.si.getFdnoByName ("default")
	
	def delete (self, qs):
		self.searcher.do_delete (qs)
		
	def close (self):
		if self.reader:
			memory.remove (self.reader)
			self.reader.close ()
			self.reader = None
		if self.searcher:
			self.searcher.close ()
	
	def getN (self):
		return self.si.segments.N
				
	def getDF (self, mem, term):
		try: 
			return self.DFCache [term]
		except KeyError:
			value = self.searcher.segments.last ().getTermInfo (mem, term, self.fdnoterm)
			if value:
				df = value [0]
				self.DFCache [term] = df
				return df
		return 0		
	
	def getIDF (self, df):
		if not self.IDFCache:
			for i in range (1, 128):
				self.IDFCache [i] = math.log (float (self.getN()) / i)
		try:
			return self.IDFCache [df]
		except KeyError:
			return math.log (float (self.getN()) / df)
	
	def getCoOccurrence (self, mem, x, y):
		if x > y:
			t = x; x = y; y = t			
		if x == y:
			raise ValueError("Should x != y")
		seekp = x * 8
		pos, df = self.reader.ci.listgetat (mem, seekp, 4, 2)
		conoccur = self.reader.co.getbyid (mem, y, df, pos, 2)
		return conoccur 
		
	def readTermInfo (self, mem, term):
		ti = self.reader.ti.get (mem, term, 1)
		if ti:
			df, doff, poff, skip, plen = ti
			self.reader.tf.read (mem, df, doff, skip)
			return 1
		return 0	
		
	def getFeatures (self, mem, terms, numvoca, top = 0):
		termsfv = []
		if top:
			classifier = _delune.Classifier (mem, self.corpus, numvoca, self.getN())
			for term, tf in terms:
				try: 
					score = self.termScoreCache [term]
					termsfv.append ((term, tf, score))
				except KeyError:
					if self.readTermInfo (mem, term):
						classifier.load ()
						score = classifier.scorefeature ()
						termsfv.append ((term, tf, score))
						self.termScoreCache [term] = score
			classifier.close ()
			termsfv.sort (key = lambda x: x [2], reverse = True)
			termsfv = [(x [0], x [1]) for x in termsfv [:top]]
			
		else:
			for term, tf in terms:
				if self.getDF (mem, term):
					termsfv.append ((term, tf))
		return termsfv
	


from datetime import datetime
from . import queryparser
from . import querylexer
import types
import copy
import re
import time
from .segment import fieldreader, typeinfo
from .. import _delune, memory

def makeHighLightRE (highlights):
	highlights = "|".join (highlights)
	new_highlights = ""
	for c in highlights:
		if c in '.?+*[]()\\^${}':
			new_highlights += "\\" + c
		else:
			new_highlights += c
			
	return new_highlights
	

class TermInfos:
	def __init__ (self, segments, analyzer, tfidf):
		self.segments = segments
		self.analyzer = analyzer
		self.tfidf = tfidf
		self.scorable = 0
		self.segTi = {}
		self.termDF = {}
		self.coord = {}
	
	def setTi (self, fdreader, term, lang):		
		try:
			if term [0] == "`": 
				raise ValueError # default field type
			name, value = term.split (":", 1)
		except ValueError:
			name, value = "default", term
		
		type, num = self.segments.si.getFieldInfo (name)
		fdi = fdreader (type, term, lang, name, value, num)
		if fdi.type == "coord":
			self.coord [fdi.name] = fdi
			if len (self.coord) >1: raise ValueError			
		return fdi	
	
	def get_terminfo (self, fdreader, segment, expression, lang, dup):
		if type (expression) is str: # only 1 term
			op, lop, rop = "*", expression, None
		else:
			op, lop, rop = expression
		
		if type (lop) is list:
			lop = self.get_terminfo (fdreader, segment, lop, lang, dup)
			
		if type (rop) is list:
			rop = self.get_terminfo (fdreader, segment, rop, lang, dup)
		
		for each in (lop, rop):
			if not each: 
				continue
				
			if each not in self.segTi [fdreader.segment]:			
				ti = self.setTi (fdreader, each, lang)				
				self.segTi [fdreader.segment][each] = ti				
			else:
				ti = self.segTi [fdreader.segment][each]
			
			if self.tfidf and typeinfo.typemap.isscorable (ti.type):
				self.scorable = 1
				if type (ti.value) is str:
					terms = [ti]
				else:
					terms = ti.value
				
				for each in terms:
					k = (each.value, each.num)
					if k in dup: 
						dup [k] += 1
						continue
												
					try:
						self.termDF [k] += each.df
					except KeyError:
						self.termDF [k] = each.df
						
					dup [k] = 1
				
	def collect (self, expression, lang):
		for segment in self.segments:
			if segment.numDoc == 0:
				continue
			
			try:
				fdreader = fieldreader.FieldReader (memory.get (segment), segment, self.analyzer, self.tfidf)
			except KeyError:
				# no field name
				continue
					
			self.segTi [fdreader.segment] = {}
			dup = {}
			self.get_terminfo (fdreader, segment, expression, lang, dup)			
			
	def isScorable (self):
		return self.scorable
				
	def getTi (self, segment, expression):
		return self.segTi [segment][expression]
	
	def getDF (self, ti):
		try:
			return self.termDF [(ti.value, ti.num)]
		except KeyError:
			return 0		
	
	def getCoord (self):		
		try:
			return list(self.coord.values ()) [0]
		except IndexError:
			return None	
		
	
class Query:
	DEBUG = 0
	CACHEPAGE = 3
	def __init__ (self, segments, qs, lang, offset, fetch, sort, analyzer, estTotal = False):
		self.segments = segments
		self.qs = qs
		self.lang = lang
		self.offset = offset
		self.fetch = fetch
		self.sort = sort
		self.analyzer = analyzer
		
		req_tfidf = (self.sort == "tfidf" or self.sort [1:] == "tfidf")
		self.terminfos = TermInfos (self.segments, self.analyzer, req_tfidf)
		self.estTotal = estTotal
		self.estRatio = None
		
		self.sortorder = 0
		self.sortindex = None
		self.sortfieldtype = None
		
		self.cachedfetch = self.offset + self.fetch
		if self.offset == 0 and self.fetch < 100:
			if len (self.segments) == 0:
				self.cachedfetch = 0
			elif len (self.segments) < self.CACHEPAGE:
				self.cachedfetch = self.fetch * self.CACHEPAGE
				if self.cachedfetch > 100: self.cachedfetch = 100
			else:
				self.cachedfetch = self.fetch * 2
				if self.cachedfetch > 100: self.cachedfetch = 100
		
		self.hits = _delune.Compute (memory.get (), req_tfidf and self.segments.si.getN () or 0)
		self.totalcount = 0
		self.duration = 0
		self.numanalyze = 0
		self.esegments = []
		self.bucket = []
		self.highlights = {}		
			
	def setSort (self):
		if self.sort:
			if self.sort [0] == "-":
				self.sortorder = -1
				self.sort = self.sort [1:]
			elif self.sort [0] == "+":
				self.sortorder = 1
				self.sort = self.sort [1:]		
			else:
				self.sortorder = -1
				
		if self.sort == "tfidf" and self.terminfos.isScorable ():
			#self.sortorder = -1
			self.sortindex = -1

		elif self.sort:
			type, num = self.segments.si.getFieldInfo (self.sort)
			if typeinfo.typemap.isint (type) and len (self.segments):
				sp = self.segments.first ().getSortMapPointer (num) # checking sortmap
				if sp:
					self.sortfieldtype = "int"							
					self.sortindex = -1
				else:
					self.sort = None
					
			elif type == "coord" and self.terminfos.getCoord () and len (self.segments):
				sp = self.segments.first ().getSortMapPointer (num) # checking sortmap
				if sp:
					self.sortfieldtype = "coord"
					self.sortindex = -2
				else:
					self.sort = None
			else:
				self.sort = None
							
	def getBufferSize (self, segment, size):
		hits = self.hits.count ()
		if hits == 0: return 0
		bsize = int ((self.segments.si.getSegmentNumDoc (segment.seg) / float (hits)) * size)
		if bsize > memory.get_buffer_size (): bufsize = 1024
		else: bufsize = 0
		return bufsize

	def close (self):
		del self.terminfos
		self.hits.close ()

	def setResult (self, cached):
		self.totalcount, highlights, self.bucket = cached
		for each in highlights:
			self.highlights [each] = None

	def newSegmentScan (self, segment):
		self.randomscan = 0
		self.numvaildqueryterm = 0
		self.numstopword = 0
		self.numanalyze = 0
		self.mem = memory.get (segment)
		self.hits.newscan ()

	def addHighlight (self, term):
		self.highlights [term] = None
		if len (term) < 3: return
		if ord (term [0]) >= 132: return
		
		# plural
		if term [-1] != 's':
			if term [-1] in "xz":
				self.highlights [term + 'es'] = None				
			elif term [-1] == 'y':
				if term [-2] in "aeiou":
					self.highlights [term + "s"] = None
				else:
					self.highlights [term [:-1] + "ies"] = None
			elif term [-1] == 'f':
				self.highlights [term [:-1] + "ves"] = None
				self.highlights [term + "s"] = None
			elif term [-2:] == 'fe':
				self.highlights [term [:-2] + "ves"] = None
				self.highlights [term + "s"] = None
			else:
				self.highlights [term + 's'] = None		
			
		elif len (term) >= 5 and term [-3:] == 'ies':
			self.highlights [term [:-3] + "y"] = None

		elif len (term) >= 5 and term [-3:] == 'ves':
			self.highlights [term [:-3] + "f"] = None
			self.highlights [term [:-3] + "fe"] = None

		elif len (term) >= 5 and term [-2:] == 'es' and (term [-3] in "sxz" or term [-4:-2] in ("sh", "ch")):
			self.highlights [term [:-2]] = None
			self.highlights [term [:-1]] = None

		elif term [-1] == 's':
			self.highlights [term [:-1]] = None
	
	def get_terminfo (self, segment, term):
		return self.terminfos.getTi (segment, term)
		
	def optimize (self, segment, expression):
		op = expression [0]				
		tis = [
			type (expression [1]) is str and self.get_terminfo (segment, expression [1]) or None, 
			type (expression [2]) is str and self.get_terminfo (segment, expression [2]) or None
		]
				
		#swap between operands
		if op in "*+":
			if tis [1] is None and self.hits.count ():
				tis  = [None, tis [0]] #avoid using stack
				if self.DEBUG: print("SWAP OPERAND TO RIGHT:", tis)

			# swap phrase
			elif tis [0] and tis [1]:
				if not tis [0].methology == "Phrase" and tis [1].methology == "Phrase":
					tis  = [tis [1], tis [0]]
					if self.DEBUG: print("SWAP PHRASE:", tis)

				#swap with df
				elif tis [0].methology == "TermList" and tis [1].methology == "TermList":
					if len (tis [0].value) < len (tis [1].value):
						tis = [tis [1], tis [0]]
						if self.DEBUG: print("SWAP TERMLIST:", tis)
							
					elif len (tis [0].value) and len (tis [1].value):
						if tis [0].value [0].df > tis [1].value [0].df:
							tis = [tis [1], tis [0]]
							if self.DEBUG: print("SWAP DF:", tis)
				
				#swap range
				elif typeinfo.typemap.getsize (tis [0].type) or typeinfo.typemap.getsize (tis [1].type):
					if typeinfo.typemap.getsize (tis [0].type):
						rf = tis [0]
						ff = tis [1]
					else:
						rf = tis [1]
						ff = tis [0]
					
					if not typeinfo.typemap.getsize (ff.type):						
						if ff.value and type (ff.value) is list:
							df = ff.value [0].df
						else:
							df = ff.df
						
						if df > self.segments.si.getN () / 10.0:
							tis = [rf, ff]
							if self.DEBUG: print("FORWARD RANGE:", tis)
						else:
							tis = [ff, rf]				
							if self.DEBUG: print("BACKWARD RANGE:", tis)
						
		return tis

	rx_proxdelta = re.compile ("\s*(?P<delta>\^[0-9]+)\s*")
	
	def computeTermList (self, segment, term, operator):
		if term.stopword: 
			return self.computeTerm (segment, term, operator)
			
		if not term.type: return 0
		if not term.value: return 0
		
		if len (term.value) == 1: # not term list	or phrase		
			term.value [0].readprox = 0
			return self.computeTerm (segment, term.value [0], operator)
		
		dc = 0
		self.randomscan = 0

		pushed = 0
		if self.hits.count ():
			self.hits.push ()
			pushed = 1
			if self.DEBUG: print("L-PUSH:", self.hits.saved ())
		
		for termTi in term.value:			
			if not termTi.df: break
			self.computeTerm (segment, termTi, operator)
			dc = self.hits.intersect (termTi.near, termTi.loose)
			if not dc: break

		if pushed:
			self.hits.popleft ()
			if self.DEBUG: print("L-POP:", self.hits.saved ())

		return dc
		
	def computeWildCard (self, segment, term, operator):
		if not term.ti: return 0

		self.randomscan = 0
		pushed = 0
		if self.hits.count ():
			self.hits.push ()
			pushed = 1
			if self.DEBUG: print("P-PUSH:", self.hits.saved ())

		for termTi in term.value:
			self.computeTerm (segment, termTi, operator)
			dc = self.hits.union ()

		if pushed:
			self.hits.popleft ()
			if self.DEBUG: print("P-POP:", self.hits.saved ())

		return dc

	def computeTerm (self, segment, term, operator):
		if isinstance (term, fieldreader.Term) or isinstance (term, fieldreader.TermList):
			self.numanalyze += 1
			if term.stopword:
				self.numstopword += 1
				return -1

		if not term.type: return 0
		if not term.df: return 0

		if self.randomscan:
			if len (self.bucket) >= self.cachedfetch:
				if self.DEBUG: print("SKIP: randomscan")
				self.hits.setcount (term.df)
				return term.df

			elif segment.rd:
				self.randomscan = 0

		df, doff, poff, skip, plen = term.ti
		if (term.sp): segment.sm.read (self.mem, self.getBufferSize (segment, term.sp [1]), *term.sp) # needn't for f type

		if self.randomscan:
			if self.DEBUG: print("RANDOM SCAN")
			segment.readPosting (self.mem, df, doff, poff, skip, plen, 0, self.cachedfetch, term.readprox)
			self.hits.setcount (term.df)

		else:
			if self.DEBUG: print("FULL SCAN")
			segment.readPosting (self.mem, df, doff, poff, skip, plen, -1, -1, term.readprox)
		
		if operator != "-" and typeinfo.typemap.hasnorm (term.type):
			if term.qstr:
				for hk in term.qstr.split ():
					self.addHighlight (hk)
					
		t =  self.hits.set (self.terminfos.getDF (term), self.segments.si.getWeight (term.name))		
		return t

	def computeCoord (self, segment, term, operator):
		if not term.value: return 0
		latitude, longitude, distance = term.value
		
		segment.sm.read (self.mem, self.getBufferSize (segment, term.sp [1]), *term.sp)
		return self.hits.distance (latitude, longitude, distance, operator.encode ("utf8"), 1)

	def computeDigit (self, segment, term, operator):
		if not term.value: return 0

		a, b = term.value
		if b in ("all", "any", "none"):
			smfunc = self.hits.bit
		else:
			smfunc = self.hits.between
		
		segment.sm.read (self.mem, self.getBufferSize (segment, term.sp [1]), *term.sp)

		if not (a == -1 and (b == -1 or b in ("all", "any", "none"))):
			self.randomscan = 0

		if self.randomscan:
			wholes = segment.getNumDoc ()  - segment.getDeletedCount ()
			self.hits.setcount (wholes)

			if len (self.bucket) >= self.cachedfetch:
				return wholes
			
			dc = smfunc (a, b, operator.encode ("utf8"), (b not in ("all", "any", "none") and self.sort == term.name) and 1 or 0, self.cachedfetch + segment.getDeletedCount ())
			return wholes

		else:
			return smfunc (a, b, operator.encode ("utf8"), (b not in ("all", "any", "none") and self.sort == term.name) and 1 or 0, -1)

	def compute (self, segment, expression):		
		if self.DEBUG: print("EXPRESSION", expression)
		record_count = 0
		got_stopwords = 0
		#---------------------------------------------------------------------------------
		# Single Query Term
		#---------------------------------------------------------------------------------
		if type (expression) is str: # only 1 term
			ti = self.get_terminfo (segment, expression)			
			record_count = getattr (self, "compute" + ti.methology) (segment, ti, "*")
			if self.DEBUG: print("SINGLE QUERY:", ti, record_count)
			return None
		
		#---------------------------------------------------------------------------------
		# Optimize inner query for saving stack
		#---------------------------------------------------------------------------------
		if expression [0] != "-" and type (expression [1]) == str and type (expression [2]) == list:
			op, lop, rop = expression [0], expression [2], expression [1]
		else:
			op, lop, rop = expression
		
		#---------------------------------------------------------------------------------
		# Recursive Compute
		#---------------------------------------------------------------------------------
		if type (lop) is list:
			lop = self.compute (segment, lop)
			
			if self.numvaildqueryterm and self.hits.count () == 0 and op in "*-":
				if self.DEBUG: print("ABORT")
				self.hits.abort () # need not proceed
				return None

		if type (rop) is list:
			rop = self.compute (segment, rop)
		
		#---------------------------------------------------------------------------------
		# Optimize
		#---------------------------------------------------------------------------------
		lop, rop = self.optimize (segment, [op, lop, rop])
		
		#---------------------------------------------------------------------------------
		# Left Operand
		#---------------------------------------------------------------------------------
		if lop:
			if self.hits.count ():
				self.hits.push ()
				if self.DEBUG: print("PUSH:", self.hits.saved ())
	
			record_count = getattr (self, "compute" + lop.methology) (segment, lop, "*") # left operator is always "*"
			if record_count != -1: self.numvaildqueryterm += 1
			if self.DEBUG: print("LQUERY:", lop, record_count)
			if not record_count	and op in "-*":
				if self.DEBUG: print("ABORT")
				self.hits.abort () #need not proceed
				return None
			
			elif record_count == -1: #stop word
				if op == "-":
					if self.DEBUG: print("ABORT (- op. but stopword)")
					self.hits.abort ()
					return None
				got_stopwords = 1

		#---------------------------------------------------------------------------------
		# Right Operand
		#---------------------------------------------------------------------------------
		if rop:
			record_count = getattr (self, "compute" + rop.methology) (segment, rop, op)
			if self.DEBUG: print("RQUERY:", rop, record_count)
			if record_count != -1: self.numvaildqueryterm += 1
			if not record_count:
				if op in "*": # + or - is keep hits
					if self.DEBUG: print("ABORT")
					self.hits.abort ()
					return
			
			elif record_count == -1:				
				got_stopwords = 1
			
			if rop.methology in ("Coord", "Digit"): # already quick operated.
				return None
		
		elif self.hits.saved ():
			if lop is None:
				self.hits.popleft ()
				if self.DEBUG: print("POPLEFT:", self.hits.saved ())
		
			else:
				self.hits.popright ()
				if self.DEBUG: print("POPRIGHT:", self.hits.saved ())
		
		#---------------------------------------------------------------------------------
		# Operate
		#---------------------------------------------------------------------------------
		if got_stopwords: # not need operation
			return None
		
		if self.DEBUG:
			print("START OPERATE:", op, lop, rop)
				
		if op == "*":
			record_count = self.hits.intersect ()
		elif op == "+":
			record_count = self.hits.union ()
		elif op == "-":				
			record_count = self.hits.trim ()

		if self.DEBUG: print("OPERATED:", record_count)
		return None
		
	def merge (self):
		if self.sortindex:
			if self.sortorder == 1:
				self.bucket.sort (key = lambda x: x[self.sortindex])				
			elif self.sortorder == -1:
				self.bucket.sort (key = lambda x: x[self.sortindex], reverse = True)				

		highlights = list(self.highlights.keys ())
		highlights.sort (key = lambda x: len (x), reverse = True)		
		#rx_highlights  = re.compile ("(^| )(?P<match>" + "|".join (highlights) + ")($| )", re.I)

	def query (self):
		s = time.time ()		
		if not len (self.segments): return			
		try:
			tokens = querylexer.QueryLexer (self.qs).tokenize ()			
			expression = queryparser.ArithmeticParser (tokens).parse_expression()			
			# collecting all term infos
			self.terminfos.collect (expression, self.lang)			
		except:
			raise			
		
		# set sorting options
		self.setSort ()
		
		for segment in self.segments:
			if segment.numDoc == 0:
				continue
			try:
				self.newSegmentScan (segment)
				# 1 term and no sort and has freq data
				if not self.sort and type (expression) is str:
					self.randomscan = 1
				self.query_segment (segment, expression)
				
			except IOError:
				#self.esegments.append (segment.seg)
				self.segments.si.trace ()
		
		if self.DEBUG: print("ANALYZED: %d STOPWORD: %d" % (self.numanalyze, self.numstopword))
		if self.numanalyze and self.numstopword == self.numanalyze:
			self.bucket = []
			self.totalcount = 0
			return

		if self.DEBUG: print("Total Result:", self.totalcount)

		self.duration = int ((time.time () - s) * 1000) #ms
		self.merge ()
			
	def query_segment (self, segment, expression):
		#-------------------------------------------------------------------------------
		# skip extra fetching when no sort with estimate-total option
		#-------------------------------------------------------------------------------
		if not self.sort and len (self.bucket) >= self.cachedfetch and self.estRatio is not None:
			self.totalcount += int (segment.getNumDoc () * self.estRatio)
			return
			
		#-------------------------------------------------------------------------------
		# operate
		#-------------------------------------------------------------------------------
		self.compute (segment, expression)
		
		#-------------------------------------------------------------------------------
		# calculating estimate ratio
		#-------------------------------------------------------------------------------
		if self.estTotal:
			if self.estRatio is None:
				self.estRatio = float (self.hits.count ()) / segment.getNumDoc ()
			else:
				self.estRatio = (self.estRatio + (float (self.hits.count ()) / segment.getNumDoc ())) / 2.
		
		#-------------------------------------------------------------------------------
		# skip extra fetching when no sort
		#-------------------------------------------------------------------------------
		if not self.sort and len (self.bucket) >= self.cachedfetch:
			count = self.hits.count ()
			self.totalcount += count
			return
		
		#-------------------------------------------------------------------------------
		# forcely coordination setting
		#-------------------------------------------------------------------------------
		if self.terminfos.getCoord ():
			coord = self.terminfos.getCoord ()			
			type, num = self.segments.si.getFieldInfo (coord.name)
			sp = segment.getSortMapPointer (num) # checking sortmap
			if sp:
				pointer, vsize = sp
				segment.sm.read (self.mem, self.getBufferSize (segment, vsize), pointer, vsize)
				self.hits.sortset_coord (coord.value [0], coord.value [1])
		
		#-------------------------------------------------------------------------------
		# sorting
		#-------------------------------------------------------------------------------
		bysortkey = 0
		byextra = 0
		if self.sort and self.sort != "tfidf":
			if self.sortfieldtype == "int":
				type, num = self.segments.si.getFieldInfo (self.sort)
				sp = segment.getSortMapPointer (num) # checking sortmap			
				if sp:
					pointer, vsize = sp
					segment.sm.read (self.mem, self.getBufferSize (segment, vsize), pointer, vsize)
					self.hits.sortset_key ()
					bysortkey = 1
										
			elif self.sortfieldtype == "coord":				
				byextra = 1
		
		want = self.cachedfetch
		self.hits.sort (want, self.sortorder, bysortkey, byextra)

		#-------------------------------------------------------------------------------
		# select from hit docs
		#-------------------------------------------------------------------------------
		bucket = self.hits.hitdoc (segment.seg, want, segment.getBits ())
		self.bucket +=  bucket

		#-------------------------------------------------------------------------------
		# counting hitdoc
		#-------------------------------------------------------------------------------
		count = self.hits.count ()
		self.totalcount += count

		if self.DEBUG: 
			print("#" * 79)
			print("Segment Result:", count)
			print("#" * 79)



# 2014. 12. 9 by Hans Roh hansroh@gmail.com

__version__ = "0.3.1.2"
version_info = tuple (map (lambda x: not x.isdigit () and x or int (x),  __version__.split (".")))

from .analyzers import standardAnalyzer
from .searcher import collection
from .classifier import model
from . import memory
from . import analyzers
import platform
import os
from .searcher import indexer, document
from .classifier import trainer, labeld_document
import threading

PID = None

standard_analyzer = standardAnalyzer.Analyzer

collection = collection.Collection
model = model.Model

document = document.Document
labeled_document = labeld_document.LabeledDocument

# Searchable Field Types
TEXT = "Text"
TERM = "Term"
STRING = "String"
LIST = "List"
FNUM = "Fnum"
COORD = "Coord4" 
COORD4 = "Coord4" # 4 decimals, 10 M precision
COORD6 = "Coord6" # 6 decimals, 10 CM precision
COORD8 = "Coord8" # 8 decimals, 1 MM precision
BIT8  = "Bit8"
BIT16 = "Bit16"
BIT24 = "Bit24"
BIT32 = "Bit32"
BIT40 = "Bit40"
BIT48 = "Bit48"
BIT56 = "Bit56"
BIT64 = "Bit64"
INT8  = "Int8"
INT16 = "Int16"
INT24 = "Int24"
INT32 = "Int32"
INT40 = "Int40"
INT48 = "Int48"
INT56 = "Int56"
INT64 = "Int64"

# Both Classifiers & Term Clustering
ALL = "default"

# Classifiers
NAIVEBAYES = "naivebayes"
FEATUREVOTE = "featurevote"
TFIDF = "tfidf"
SIMILARITY = "similarity"
ROCCHIO = "rocchino"
ROCCHINO = "rocchino"
MULTIPATH = "multipath"
META = "metaclassifier"

# Term Clustering
TERMCLUSTER = "termcluster"

# Feature Scoring
CHI2 = "chi"
GSS = "gss"
DF = "df"
CF = "cf"
NGL = "ngl"
MI = "mi"
TFIDF = "tfidf"
IG = "ig"
OR = "or"
OR4P = "or4p"
RS = "rs"
LOR = "lor"
COS = "cos"
PPHI = "pphi"
YULE = "yule"
RMI = "rmi"

# Feature Scoring Used For Term Clustring Only
R = "r"

# Scoring Value
SUM = "sum"
MAX = "max"
AVG = "avg"
MIN = "min"

# Open Modes
WRITE = "w"
APPEND = "w"
MODIFY = "w"
READ = "r"
CREATE = "c"

osbit, __x = platform.architecture()
if osbit == "64bit":
	LIMIT_SEGMENTSIZE = 10000000000	# 2**39, 10GB but max 500GB is possible
else:
	LIMIT_SEGMENTSIZE = 1900000000 # 2**31, 2GB
LIMIT_FILESIZE = int (LIMIT_SEGMENTSIZE * 0.7)

logger = None
	
class Task:
	def __init__ (self, logger):
		self.logger = logger
		self.lock = threading.RLock ()
		self._d_ = {}
		
	def cleanup (self):
		with self.lock:
			reactors = self._d_.items ()
		for alias, reactor in reactors:
			self.logger ("...delune closing %s" % alias, "info")
			reactor.close ()
			
	def __contains__ (self, alias):
		with self.lock:
			return (alias in self._d_)
	
	def __getattr__ (self, alias):
		return self.get (alias)
	
	def __delattr__ (self, alias):		
		self.resign (alias)

	def __dir__ (self):
		with self.lock:
			return list (self._d_.keys ())
	
	def __len__ (self):
		with self.lock:
			return len (self._d_)
	
	def swap (self, a, b):
		with self.lock:
			t = self._d__ [a]
			self._d__ [a] = b
			self._d__ [b] = t
		
	def assign (self, alias, obj):
		obj.si.set_ident (alias)
		with self.lock:
			self._d_ [alias] = obj
	
	def resign (self, alias):
		with self.lock:
			del self._d_ [alias]
		
	def get (self, alias):
		with self.lock:
			return self._d_.get (alias)
	
	def status (self):
		d = {}
		with self.lock:
			for alias, reactor in self._d_.items ():
				d [alias] = reactor.status ()
		return d
		

class NoTask:
	def __getattr__ (self, name):
		raise AssertionError ("delune not configured, configure first")

task = NoTask ()

def configure (numthread, logger_t, io_buf_size = 16384, mem_limit = 128, max_segment_size = 0):
	global logger, task, PID
	
	if isinstance (task, Task):
		return
	
	PID = os.getpid ()		
	logger = logger_t
	if max_segment_size:
		set_max_segment_size (max_segment_size)
		
	if not memory.isInitialized ():
		memory.initialize (numthread, io_buf_size, mem_limit, "segment", logger_t)
		analyzers.buildFactory (numthread, logger_t)
	
	task = Task (logger)
	return task
	
def shutdown ():
	global task
	
	if task is None:
		return
		
	if isinstance (task, Task):
		task.cleanup ()
		
	if memory.isInitialized ():
		analyzers.close ()
		memory.destroy ()
	
	task = None	

# For Skitai
cleanup = shutdown

def set_max_segment_size (mbytes):
	global LIMIT_SEGMENTSIZE, LIMIT_FILESIZE
	LIMIT_SEGMENTSIZE = int (mbytes * 1000000)
	LIMIT_FILESIZE = int (LIMIT_SEGMENTSIZE * 0.7)

def qualify_analyzer (analyzer):
	if memory.isInitialized ():
		return analyzers.checkIn (analyzer) 
	return analyzer	


#=========================================
# query and guess by task alias
#=========================================

#--------------------
# common jobs
#--------------------
	
def assign (alias, obj):
	global task
	task.assign (alias, obj)
	
def resign (alias):
	global task
	task.resign (alias, obj)
		
def close (alias, *args, **karg):
	global task	
	reactor = task.get (alias)
	if not reactor.closed:
		reactor.close (*args, **karg)
	task.resign (alias)

def drop (alias):
	try:
		stu = status (alias)
	except KeyError:
		return
	close (alias)	
	for data_dir in stu ['indexdirs']:
		for fn in os.listdir (data_dir):
			os.remove (os.path.join (data_dir, fn))
	
def get (alias):
	global task
	return task.get (alias)

def status (alias = "", *args, **karg):
	global task
	if alias == "":
		return task.status ()
	return task.get (alias).status (*args, **karg)

def refresh (alias, *args, **karg):
	global task
	return task.get (alias).refresh (*args, **karg)	

#--------------------
# searcher only jobs	
#--------------------
def query (alias, *args, **karg):
	global task
	return task.get (alias).query (*args, **karg)

def delete (alias, *args, **karg):
	global task
	return task.get (alias).delete (*args, **karg)		

def remove (alias, *args, **karg):
	global task
	return task.get (alias).remove (*args, **karg)
	
def fetch (alias, *args, **karg):
	global task
	return task.get (alias).fetch (*args, **karg)	

#----------------------
# classifier only jobs	
#----------------------
def guess (alias, *args, **karg):
	global task
	return task.get (alias).guess (*args, **karg)	



try:
	import pycld2 as cld2
	
except ImportError:
	def detect (text):
		raise SystemError ("pycld2 is not installed, use pip install pycld2")
		
else:	
	def detect (text):
		detected = cld2.detect (text)	
		if detected [0] is False:
			return None
		else:
			return detected [2][0][1]




import os
import re
try:
	import urllib.parse
	unquote = urllib.parse.unquote
except ImportError:
	import urlparse
	unquote = urlparse.unquote
		
import types
import sys

def mkdir (tdir, mod = -1):
	while tdir:
		if tdir [-1] in ("\\/"):
			tdir = tdir [:-1]
		else:
			break	

	if os.path.isdir (tdir): return	
	chain = [tdir]	
	while 1:
		tdir, last = os.path.split (tdir)				
		if not last: 
			break
		if tdir:
			chain.insert (0, tdir)
	
	for dir in chain [1:]:
		try: 
			os.mkdir (dir)
			if os.name == "posix" and mod != -1:
				os.chmod (dir, mod)				
		except OSError as why:
			if why.errno in (17, 183): continue
			else: raise


def modpath (mod_name):	
	if type (mod_name) in (str, bytes):		
		try:
				mod = sys.modules [mod_name]
		except KeyError: 
			return "", ""		
		return mod.__name__, mod.__file__

			
NAFN = re.compile (r"[\\/:*?\"<>|]+")
def mkfn (text):
	text = unquote (text)
	return NAFN.sub ("_", text)




from . import segment
from . import segmentreader
from ... import _delune
import time

class SegmentMerger (segment.Segment):
	def __init__ (self, si, truncate = False):
		self.si = si
		self.truncate = truncate
		self.delay_sec, self.delay_factor = self.si.getopt (lazy_merge = (0, 3))
		if not self.delay_factor:
			self.delay_factor = 3
		newseg = self.si.getNewSegment ()
		segment.Segment.__init__ (self, si.fs.new (newseg), newseg, 'w', version = self.si.version)
		self.mergeinfo = _delune.MergeInfo ()
		self.smis = self.mergeinfo.get ()
		self.segments = []
		self.deletables = []
	
	def close (self):
		if self.mergeinfo:
			self.mergeinfo.close ()
			self.mergeinfo = None
		segment.Segment.close (self)
			
	def addSegment (self, seg):
		reader = segmentreader.SegmentBulkReader (self.si, seg)
		self.deletables.append (reader)
		if reader.numDoc and not self.truncate:
			self.segments.append (reader)
		
	def merge (self):
		self.segments.sort (key = lambda x: int (x.seg))
		
		for segment in self.segments:
			self.mergeinfo.add (*segment.getMergeInfo ())
			
		self.tf.setsmis (self.smis)	
		self.fd.setsmis (self.smis)
		self.sm.setsmis (self.smis)
		self.si.log ('merging documents...')
		self.mergeDocument ()		
		self.si.log ('merging sort maps...')
		self.mergeSortMap ()		
		self.si.log ('merging sort terms and postings...')
		self.mergeTermInfo ()
		self.flush ()
		self.si.log ('merging finished')
		
	def flush (self):
		for reader in self.deletables:
			seg = reader.seg
			reader.close ()
			self.si.removeSegment (seg)
						
		self.si.addSegment (self.seg, self.numDoc)
		self.close ()
		self.si.flush ()
		
	def mergeTermInfo (self):
		delay_factor = self.delay_factor * 10000
		slots = [None] * len (self.segments)
		i = 0
		for segment in self.segments:
			slots [i] = segment.advanceTermInfo ()
			i += 1			
		
		#slots = slots [:1]
		termnum = 0
		while 1:
			fslots = [_f for _f in slots if _f]
			if not fslots: break
			#slots = slots [:1]
			
			miti = min (fslots)			
			
			_term, _fdno = miti.term, miti.fdno			
			i = 0
			last_docid = 0
			tdf = 0
			frqposition, prxposition = self.tf.tell ()
			for ti in slots:
				if ti and miti == ti:
					tdf, skip, pskip, last_docid = self.tf.merge (i, ti.df, ti.frqPosition, ti.prxPosition, ti.prxLength, tdf, last_docid)
					try: 
						slots [i] = self.segments [i].advanceTermInfo ()					
					except IndexError: 
						slots [i] = None
				i += 1
			
			# only when has docments
			if not tdf:
				continue
				
			self.ti.add (_term, _fdno, tdf, frqposition, prxposition, skip - frqposition, pskip - prxposition) #skipdelta: skip - position
			self.tf.commit ()			
			termnum += 1
			if termnum and termnum % delay_factor == 0:
				time.sleep (self.delay_sec)
			if termnum and termnum % 500000 == 0:
				self.si.log ('..%dK terms' % (termnum / 1000))
			
	def mergeDocument (self):
		delay_factor = self.delay_factor * 1000
		for i in range (len (self.segments)):
			self.si.log ('..segment %s' % self.segments [i].seg)
			for docid in range (self.segments [i].numDoc):				
				self.numDoc += self.fd.merge (i, docid)
				if self.numDoc and self.numDoc % delay_factor == 0:
					time.sleep (self.delay_sec)
			
	def mergeSortMap (self):
		sminfos = self.si.getSortMapFields ()
		works = 0
		for fdno, _size in sminfos:
			works += 1
			self.si.log ('..%d of %d' % (works, len (sminfos)))
			for i in range (len (self.segments)):				
				pointer, size = self.segments [i].getSortMapPointer (fdno)				
				self.sm.merge (i, fdno, pointer, size)
			time.sleep (self.delay_sec)
			


import time
import random
import psyco
psyco.full ()

class IOBench:
	def __init__ (self):
		self.f = open ("c:\\.corpus", "rb")
		self.pointers = [random.randrange (13000000) for i in range (1024 * 1024 / 8)]
		
	def close (self):
		self.f.close ()
		
	def read (self, kb, num = 3):
		read = 0
		readsize = 1024 * kb
		loop = 1024 * 1024 / kb
		t = 0
		for i in range (num):
			s = time.time ()
			for k in range (loop):
				self.f.seek (self.pointers [k])
				data = self.f.read (readsize)
				read += len (data)
			t += time.time () - s
		
		print("file reading %d bytes, per each %d Kbytes reading, %2.3f seconds" % (read, kb, t / num))


f = IOBench ()

f.read (512, 3)
f.read (256, 3)
f.read (128, 3)
f.read (64, 3)
f.read (32, 3)
f.read (16, 3)
f.read (8, 3)
f.read (8, 3)
f.read (16, 3)
f.read (32, 3)
f.read (64, 3)
f.read (128, 3)
f.read (256, 3)
f.read (512, 3)

f.close ()



"""
Experimental Classifier
2016.4.6
"""

from . import metaclassifier, termcluster
import delune

class Classifier (metaclassifier.Classifier):
	def __init__ (self, reader, analyzer, featset, use_top_features, logger = None):
		self.subcl = delune.NAIVEBAYES
		self.cluster = None
		metaclassifier.Classifier.__init__ (self, reader, analyzer, featset, use_top_features, logger)
		
	def setopt (self, **opts):	
		self.subcl = opts.get ("subcl", self.subcl)
		self.cluster = opts.get ("cluster", self.cluster)
		metaclassifier.Classifier.setopt (self, **opts)
		if self.cluster:
			self.cluster.setopt (**opts)

	def match (self, name):
		return name == "multipath"
		
	def guess (self, mem, qs, lang = "un", cond = ""):
		terms = self.cluster.getFeatures (mem, qs, lang)
		clusters = self.cluster.cluster (mem, qs, lang, terms)
		
		results = []
		for cluster in clusters:
			if self.verbose: self.logger ("cluster contains %d of %d terms" % (len (cluster), len (terms)))
			result = self.clfs [self.subcl].guess (mem, terms, lang, cond)			
			results.append (result)
			
		return metaclassifier.Classifier.merge_and_choice (self, results)
		




import threading
import time

class Cache:
	def __init__ (self, max = 30):
		self.max = max
		self.cache = {}
		self.timeused = {}
		self.mutex = threading.Lock ()

	def clear (self):
		self.mutex.acquire ()
		self.cache.clear ()
		self.timeused.clear ()
		self.mutex.release ()

	def set (self, key, value):
		if self.max <= 0: return
		self.mutex.acquire ()
		if len (self.cache) > self.max:
			self.drop ()

		self.cache [key] = value
		self.timeused [key] = [0, int (time.time ())]
		self.mutex.release ()

	def get (self, key):
		self.mutex.acquire ()
		try:
			value = self.cache [key]
			t, u = self.timeused [key]
		except KeyError:
			self.mutex.release ()
			return None

		self.timeused [key] = [u + 1, int (time.time ())]
		self.mutex.release ()
		return value

	def remove (self, key):
		try: del self.cache [key]
		except KeyError: pass
		try: del self.timeused [key]
		except KeyError: pass

	def drop (self):
		minkey = None
		lasttu = None
		for key, tu in list(self.timeused.items ()):
			if minkey is None:
				minkey = key
				lasttu = tu
			elif tu < lasttu:
				minkey = key
				lasttu = tu

		self.remove (minkey)



import math

# standard

typeinfo = {	
	"Text": {"method": "Text"},
	"Term": {"method": "Text"},
	
	"String": {"method": "Field"},
	"List": {"method": "List"},
	"Fnum": {"method": "Fnum"},
	
	"Coord4": {"len": 6, "method": "Coord"},
	"Coord6": {"len": 8, "method": "Coord"},
	"Coord8": {"len": 10, "method": "Coord"},
	
	"Bit8":  {"len": 1, "method": "Bit"},
	"Bit16": {"len": 2, "method": "Bit"},
	"Bit24": {"len": 3, "method": "Bit"},
	"Bit32": {"len": 4, "method": "Bit"},
	"Bit40": {"len": 5, "method": "Bit"},
	"Bit48": {"len": 6, "method": "Bit"},
	"Bit56": {"len": 7, "method": "Bit"},
	"Bit64": {"len": 8, "method": "Bit"},
	
	"Int8":  {"len": 1, "method": "Int"},
	"Int16": {"len": 2, "method": "Int"},
	"Int24": {"len": 3, "method": "Int"},
	"Int32": {"len": 4, "method": "Int"},
	"Int40": {"len": 5, "method": "Int"},
	"Int48": {"len": 6, "method": "Int"},
	"Int56": {"len": 7, "method": "Int"},
	"Int64": {"len": 8, "method": "Int"}
}
	
class FieldType:
	def hasnorm (self, type):
		return type in ("Term", "Text")

	def isscorable (self, type):
		return self.hasnorm (type) or self.isfield (type) or self.islist (type)

	def isfield (self, type):
		return type == "String"
		
	def islist (self, type):
		return type == "List"
		
	def isbit (self, type):
		return type [:3] == "Bit"
	
	def isint (self, type):
		return type [:3] == "Int"
	
	def isextra (self, type):
		return type [:5] == "Coord"
	
	def iscoord (self, type):
		return type [:5] == "Coord"
			
	def valid (self, type):
		return type in typeinfo
	
	def getdelim (self):
		try:
			return typeinfo [type]["delim"]
		except KeyError:
			return
		
	def getsize (self, type):
		try:
			return typeinfo [type]["len"]
		except KeyError:
			return 0
	
	def getmethod (self, type):	
		if type.startswith ("Fnum"):
				return "Fnum"
		return typeinfo [type]["method"]
		
typemap = FieldType ()

def zfill (ftype, value):
	if type (value) is str:
		if value.find (".") != -1:
			try:
				value = float (value)
			except ValueError:	
				raise TypeError ('DOUBLEF Field allows float or int')							
					
		else:
			try:
				value = int (value)
			except ValueError:			
				raise TypeError ('DOUBLEF Field allows float or int')				
	
	fmt = "%%0%sf" % ftype[4:]
	val = fmt % value
	if val [0] == "-":
		val = "-0" + val [1:]
	return val		
	


from delune import _delune

class HtmlStripHandler:
	#MULTILINE_TAGS = ("script", "style", "object")
	NEED_SPACE = ("p", "br", "td", "table", "div")
	def get_parsed (self):
		return " ".join (self.buffer)
	
	#def handle_sdata (self, data):
		#print (repr (data))
			
	def reset (self):
		self.in_multiline = False
		self.last_commit = 0
		self.buffer = []
	
	def finish_starttag (self, tag, attr):
		tag = tag.lower ()
		#if tag in self.MULTILINE_TAGS:
			#self.in_multiline = True	
		if tag in self.NEED_SPACE:
			self.handle_data (" ")
		
	#def finish_endtag (self, tag):
		#print (tag)
		#if tag in self.MULTILINE_TAGS:
		#	if not self.in_multiline:
		#		self.buffer = self.buffer [:self.last_commit]
		#	else:		
		#		self.in_multiline = False
		#	self.last_commit = len (self.buffer)
		
	def handle_data (self, data):
		#print ("+", repr (data))
		#if self.in_multiline: 
		#	return		
		data = data.strip ()
		self.buffer.append (data)



class Parser:
	def __init__ (self, handler):
		self.handler = handler
		
	def parse (self, data):			
		self.handler.reset ()
		p = _delune.SGMLParser ()
		p.register (self.handler)		
		p.parse (data)		
		p.close ()
		return self.handler.get_parsed ()
	
	def close (self):
		pass
			
		
def build_parser ():
	handler = HtmlStripHandler ()
	return Parser (handler)


if __name__ == "__main__":
	import urllib.request, urllib.parse, urllib.error
	p = build_parser ()	
	url = "http://www2.openfos.com/search/key.stk.sk_3com%20switches.htm"
	f = urllib.request.urlopen (url)
	html = f.read ()
	print(p.parse (html))
	
		
	


# 2017. 3. 13 by Hans Roh (hansroh@gmail.com)

from skitai.saddle import Saddle
import skitai
import delune
from delune.lib import pathtool
import os
import json
import codecs
import time
import shutil

app = Saddle (__name__)
app.config.numthreads = 1
app.last_maintern = time.time ()

def getdir (*d):
	return os.path.join (app.config.resource_dir, *d)

def is_json (request):
	return request.command == "post" and not request.get_header ('content-type', '').startswith ('application/x-www-form-urlencoded')

def normpath (path):
	if os.name == "nt":
		return path.replace ("/", "\\")
	return path.replace ("\\", "/")
			
def load_data (alias, numthreads, plock):
	with codecs.open (getdir ("models", ".config", alias), "r", "utf8") as f:
		colopt = json.loads (f.read ())			
		colopt ['data_dir'] = [getdir (normpath(d)) for d in colopt ['data_dir']]
	
	if 'classifier' in colopt:
		analyzer = delune.standard_analyzer (10000, numthreads, **colopt ["analyzer"])
		col = delune.model (colopt ["data_dir"], delune.READ, analyzer, plock = plock, version = colopt.get ("version", 1))
		actor = col.get_classifier (**colopt.get ('classifier', {}))
	else:
		analyzer = delune.standard_analyzer (8, numthreads, **colopt ["analyzer"])
		col = delune.collection	(colopt ["data_dir"], delune.READ, analyzer, plock = plock, version = colopt.get ("version", 1))	
		actor = col.get_searcher (**colopt.get ('searcher', {}))
		actor.create_queue ()
	delune.assign (alias, actor)

def error (response, status, errcode, errmsg = "", errstack = None):
	err = response.fault (errmsg, errcode, exc_info = errstack)
	return response (status, err)
		
#-----------------------------------------------------------------

@app.startup
def startup (wasc):
	app.config.numthreads = len (wasc.threads)
	delune.configure (app.config.numthreads, wasc.logger.get ("app"), 16384, 128)
	pathtool.mkdir (getdir ("models", ".config"))
	for alias in os.listdir (getdir ("models", ".config")):
		if alias.startswith ("-"):
			with codecs.open (getdir ("models", ".config", alias), "r", "utf8") as f:
				colopt = json.loads (f.read ())
			for d in [getdir (normpath(d)) for d in colopt ['data_dir']]:
				if os.path.isdir (d):
					shutil.rmtree (d)
			os.remove (getdir ("models", ".config", alias))
		elif alias.startswith ("#"):
			continue
		else:
			load_data (alias, app.config.numthreads, wasc.plock)
  
@app.shutdown
def shutdown (wasc):
	delune.shutdown ()

#-----------------------------------------------------------------

@app.before_request
def before_request (was):
	with app.lock:
		last_maintern = was.getlu ("delune:collection") > app.last_maintern and app.last_maintern or 0
		if last_maintern:
			app.last_maintern = time.time ()
	
	if last_maintern:
		was.log ('collection changed, maintern...')
		for alias in os.listdir (getdir ("models", ".config")):			
			if alias [0] in "#-":
				if delune.get (alias [1:]):
					delune.close (alias [1:])
			elif not delune.get (alias):						
				load_data (alias, app.config.numthreads, was.plock)			
			elif os.path.getmtime (getdir ("models", ".config", alias)) > last_maintern:
				delune.close (alias)
				load_data (alias, app.config.numthreads, was.plock)

	if was.request.args.get ('alias') and was.request.routed.__name__ != "config":
		alias = was.request.args.get ('alias')
		if not delune.get (alias):
			return error (was.response, "404 Not Found", 40401, "resource %s not exist" % alias)

#-----------------------------------------------------------------

@app.route ("/")
def index (was):
	return was.response.api ({'collections': list (delune.status ().keys ())})	

@app.route ("/<alias>", methods = ["GET", "POST", "PATCH", "DELETE", "OPTIONS"])
def config (was, alias, side_effect = ""):
	fn = getdir ("models", ".config", alias)
	if was.request.command == "get":
		if side_effect == "undo":
			for mark in "#-":
				if os.path.isfile (getdir ("models", ".config", mark + alias)):
					os.rename (
						getdir ("models", ".config", mark + alias),
						getdir ("models", ".config", alias)
					)
					was.setlu ("delune:collection")					
					return was.response ("201 Accept", was.response.api ())
			return error (was.response, "404 Not Found", 20100, "resource already commited")
		
		if not delune.get (alias):
			return error (was.response, "404 Not Found", 40401, "resource %s not exist" % alias)
			
		status = delune.status (alias)
		conf = getdir ("models", ".config", alias)
		with codecs.open (conf, "r", "utf8") as f:
			colopt = json.loads (f.read ())		
			status ['colopt'] = {
				'data': colopt,
				'mtime': 	os.path.getmtime (conf),
				'size': 	os.path.getsize (conf),
				'path': conf
			}
		return was.response.api (status)
			
	if was.request.command == "delete":
		if not os.path.isfile (fn):
			return error (was.response, "404 Not Found", 40401, "resource not exist")
		
		a, b = os.path.split (fn)
		if side_effect == "data":
			newfn = os.path.join (a, "-" + b)
		else:
			newfn = os.path.join (a, "#" + b)		
		os.rename (fn, newfn)
		was.setlu ("delune:collection")
		return was.response.api ()
	
	for mark in "#-":
		if os.path.isfile (getdir ("models", ".config", mark + alias)):			
			return error (was.response, "406 Conflict", 40601, "removed resource is already exists, use UNDO")
			
	if was.request.command == "post" and delune.get (alias):
		return error (was.response, "406 Conflict", 40601, "resource already exists")		
	elif was.request.command == "patch" and not delune.get (alias):
		return error (was.response, "404 Not Found", 40401, "resource not exist")
			
	with codecs.open (fn, "w", "utf8") as f:
		f.write (was.request.body.decode ("utf8"))
	
	was.setlu ("delune:collection")		
	return was.response.api ()

#-----------------------------------------------------------------

@app.route ("/<alias>/locks", methods = ["GET", "OPTIONS"])
def locks (was, alias):	
	return was.response.api ({"locks": delune.get (alias).si.lock.locks ()})

@app.route ("/<alias>/locks/<name>", methods = ["POST", "DELETE", "OPTIONS"])
def handle_lock (was, alias, name):	
	if was.request.command == "post":
		delune.get (alias).si.lock.lock (name)		
		return was.response.api ()
	delune.get (alias).si.lock.unlock (name)
	return was.response.api ()

#-----------------------------------------------------------------

@app.route ("/<alias>/commit", methods = ["GET", "OPTIONS"])
def commit (was, alias):
	delune.get (alias).queue.commit ()
	return was.response.api ()

@app.route ("/<alias>/rollback", methods = ["GET", "OPTIONS"])
def rollback (was, alias):
	delune.get (alias).queue.rollback ()
	return was.response.api ()

#-----------------------------------------------------------------

@app.route ("/<alias>/collection/<group>/<fn>", methods = ["GET", "OPTIONS"])
def getfile (was, alias, group, fn):
	s = delune.status (alias)
	if group == "primary":
		path = os.path.join (s ["indexdirs"][0], fn)
	else:
		path = os.path.join (s ["indexdirs"][0], group, fn)
	return was.response.file (path)

@app.route ("/<alias>/segments/<group>/<fn>", methods = ["GET", "OPTIONS"])
def getsegfile (was, alias, group, fn):
	s = delune.status (alias)
	seg = fn.split (".") [0]
	if group == "primary":
		if seg not in s ["segmentsizes"]:
			return error (was.response, "404 Not Found", 40401, "resource not exist")
		path = os.path.join (s ["segmentsizes"][seg][0], fn)	
	else:
		path = os.path.join (s ["indexdirs"][0], group, fn)
	return was.response.file (path)

#-----------------------------------------------------------------

@app.route ("/<alias>/documents", methods = ["POST", "DELETE", "OPTIONS"])
def add (was, alias, truncate_confirm = ""):
	if was.request.command == "delete":
		if truncate_confirm != alias:
			return error (was.response, "400 Bad Request", 40003, 'parameter truncate_confirm=(alias name) required')
		delune.get (alias).queue.truncate ()		
		return was.response.api ()
		
	delune.get (alias).queue (0, was.request.body)
	return was.response.api ()

@app.route ("/<alias>/documents/<_id>", methods = ["GET", "DELETE", "PUT", "PATCH", "OPTIONS"])
def get (was, alias, _id, nthdoc = 0):
	if was.request.command == "get":		
		return was.response.api (delune.query (alias, "_id:" + _id, nthdoc = nthdoc))
	
	delune.get (alias).queue (1, json.dumps ({"query": {'qs': "_id:" + _id}}))
	if was.request.command in ("patch", "put"):
		delune.get (alias).queue (0, was.request.body)

	return was.response.api ()

#-----------------------------------------------------------------
	
@app.route ("/<alias>/search", methods = ["GET", "POST", "DELETE" "OPTIONS"])
def query (was, alias, **args):
	if is_json (was.request):
		args = was.request.json ()
	
	q = args.get ("q")
	if not q:
		return error (was.response, "400 Bad Request", 40003, 'parameter q required')
	
	l = args.get ("lang", "en")
	analyze = args.get ("analyze", 1)
	
	if was.request.command == "delete":
		delune.get (alias).queue (1, json.dumps ({"query": {'qs': q, 'lang': l, 'analyze': analyze}}))
		return was.response.api ()
	
	o = args.get ("offset", 0)
	f = args.get ("limit", 10)
	s = args.get ("sort", "")
	w = args.get ("snippet", 30)	
	r = args.get ("partial", "")
	d = args.get ("nthdoc", 0)	
	data = args.get ("data", 1)
	
	if type (q) is list:
		return was.response.api ([delune.query (alias, eq, o, f, s, w, r, l, d, analyze, data, limit = 1) for eq in q])
	return was.response.api (delune.query (alias, q, o, f, s, w, r, d, l, analyze, data, limit = 1))

@app.route ("/<alias>/guess", methods = ["GET", "POST", "OPTIONS"])
def guess (was, alias, **args):	
	# args: q = '', l = 'en', c = "naivebayes", top = 0, cond = ""
	if is_json (was.request):
		args = was.request.json ()
	q = args.get ("q")
	if not q:
		return error (was.response, "400 Bad Request", 40003, 'parameter q required')				
	l = args.get ("lang", 'en')
	c = args.get ("clf", 'naivebayes')
	top = args.get ("top", 0)
	cond = args.get ("cond", '')
	if type (q) is list:
		return was.response.api ([delune.guess (alias, eq, l, c, top, cond) for eq in q])
	return was.response.api (delune.guess (alias, q, l, c, top, cond))
	
@app.route ("/<alias>/cluster", methods = ["GET", "POST", "OPTIONS"])
def cluster (was, alias, **args):
	if is_json (was.request):
		args = was.request.json ()
	q = args.get ("q")
	if not q:
		return error (was.response, "400 Bad Request", 40003, 'parameter q required')
	l = args.get ("lang", 'en')	
	if type (q) is list:
		return was.response.api ([delune.cluster (alias, eq, l) for eq in q])	
	return was.response.api (delune.cluster (alias, q, l))



from . import naivebayes
from delune import _delune

class Classifier (naivebayes.Classifier):
	def match (self, name):
		return name in ("fv", "featurevote")
	
	def guess (self, mem, qs, lang = "un", cond = ""):
		if type (qs) is type ([]):
			terms = qs
		else:	
			terms = self.getFeatures (mem, qs, lang)
		if not terms: return []
		
		classifier = _delune.Classifier (mem, self.reader.corpus, self.numfeat, self.reader.getN ())
		for term, tf  in terms:			
			if self.reader.readTermInfo (mem, term):
				classifier.load ()
				classifier.featurevote (tf)

		result = classifier.get (0, self.reader.numpools)
		classifier.close ()
		return self.reader.translate (result)



from ... import _delune
from ...util import util
from . import segmentwriter
from . import fieldwriter
import types

class WDocument:
	def __init__ (self, docId, si, fw, analyzer):
		self.docId = docId
		self.si = si
		self.fw = fw
		self.analyzer = analyzer
		self.stored = []
		self.summary = ""

	def addField (self, type, name, value, lang, option = []):
		num = self.si.setFieldInfo (name, type)
		self.fw (type, self.docId, num, name, value, lang, option)

	def transport (self, text, lang, option):
		if self.analyzer and "removehtml" in option:
			temp = text
			text = self.analyzer.htmlStrip (temp)

		elif self.analyzer and "formalize" in option:
			text = self.analyzer.formalize (text)

		if self.analyzer and "transform" in option:
			text = self.analyzer.transform (text)

		# WARNING: always last processing
		if self.analyzer and "stem" in option:
			text = self.analyzer.term (text, lang)
			if text is None:
				text = []

		return text
	
	def setData (self, contents):
		self.stored = [util.serialize (content) for content in contents]		
	
	def addSummary (self, summary, lang, option = []):
		self.summary = self.transport (summary, lang, option)
		assert type (self.summary) is str


class IndexWriter:
	def __init__ (self, si, analyzer):
		self.si = si
		self.analyzer = analyzer
		self.tcache = {}

		self.fw = None
		self.segment = None
		self.newSegment ()

	def close (self):
		if self.segment:
			self.segment.close ()
		self.segment = None

	def newSegment (self):
		self.segment = segmentwriter.SegmentWriter (self.si)
		self.fw = fieldwriter.FieldWriter (self.segment, self.analyzer)

	def getNumDoc (self):
		return self.segment.numDoc

	def getMemUsage (self):
		return self.segment.getMemUsage ()

	def newWDocument (self):
		docId = self.segment.newdocid ()
		d = WDocument (docId, self.si, self.fw, self.analyzer)
		return d

	def addWDocument (self, d):
		self.segment.writeField (d.stored, d.summary)
		d = None

	#---------------------------------------------------------------------------
	# flushing
	#---------------------------------------------------------------------------
	def flush (self, final):
		if self.segment:
			self.segment.flush ()
			self.segment = None

		if not final:
			self.newSegment ()



from hashlib import md5

class AnalyzerFactory:
	def __init__ (self, numthread, logger = None):
		self.numthread = numthread
		self.logger = logger
		self.rooms = {}
	
	def close (self):	
		for a in list(self.rooms.values ()):
			a.close ()
		
	def checkIn (self, analyzer):
		l = []		
		for n in dir (analyzer):
			if n.startswith ("__"): continue
			v = str (getattr (analyzer, n))
			if v.find ("<") != -1: continue # skip instances
			l.append ("%s=%s" % (n, v))
		l.sort ()
		l.insert (0, "class=%s" % analyzer.__class__)
		sl = "&".join (l)		
		dd = md5 (sl.encode ("utf8"))
		idkey = dd.hexdigest()
		if idkey in self.rooms:
			return self.rooms [idkey]
		else:
			if len (analyzer) != self.numthread:
				analyzer.close ()
				analyzer.createTSAnalyzers (self.numthread)
			self.rooms [idkey] = analyzer
		return analyzer
		

factory = None

def buildFactory (numthread, logger):
	global factory
	factory = AnalyzerFactory (numthread, logger)

def checkIn (analyzer):
	global factory
	return factory.checkIn (analyzer)

def close ():	
	global factory
	factory.close ()
	factory = None